#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <mpi.h>

/* 
Copyright J. Russell Manson, Ph.D., P.E.
Contact me at: Russell.Manson@Stockton.edu
*/

/* Define some important variables
   that are used globally           
   Most variable names are self explanatory*/

double density_fluid = 1000.0;
double density_particle = 2650.0;
double gravity = 9.8;
double particle_diameter[1250];
double particle_mass;
double particle_volume;
double allposition[1250][3];
double allvelocity[1250][3];
double allomega[1250][3];
int icount=1;
int number_of_particles;

double effective_mass(double mass1, double mass2) 
{
/* This function is used to compute the 
   effective mass of the colliding particles. */
double mass;
return mass = mass1*mass2/(mass1+mass2);
}

double dot_product(double x[], double y[]) 
{
/* This function computes the dot product
   of two vectors.                            */
double magnitude;
return magnitude = x[0]*y[0]+x[1]*y[1]+x[2]*y[2];
}

void cross_product(double x[], double y[], double xy[]) 
{
/* This function computes the cross product
   of two vectors.                            */
xy[0] = x[1]*y[2]-x[2]*y[1];
xy[1] = -1.0*(x[0]*y[2]-x[2]*y[0]);
xy[2] = x[0]*y[1]-x[1]*y[1];
}

double norm(double x[]) 
{
/* This function computes the norm
   of a vector.                            */
double magnitude;
return magnitude = sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
}

double distance_calc(double x[], double y[]) 
{
/* This function computes the distance
   between two points given their 3d coordinates.  */
double diff[3];
double magnitude;
int i;
   for(i=0;i<3;++i)
   {
   diff[i]=y[i]-x[i];
   }
   magnitude = norm(diff);
   return magnitude;
}

void normal(double x[], double y[], double unit_normal[]) 
{
/* This function computes the unit normal vector  */
double diff[3];
double magnitude;
int i;
   for(i=0;i<3;++i)
   {
   diff[i]=y[i]-x[i];
   }
   for(i=0;i<3;++i)
   {
   unit_normal[i]=diff[i]/norm(diff);
   }
}

void tangential(double x[], double unit_tangential[]) 
{
/* This function computes the unit tangential vector  */
int i;
   for(i=0;i<3;++i)
   {
   unit_tangential[i]=x[i]/(norm(x)+1E-99);
   }
}

void collision(double mass1, double mass2, double position1[], double position2[], double velocity1_before[], double velocity2_before[], double velocity1_after[], double velocity2_after[])
{
/* Given the positions and velocities before a collision
   this function computes the velocities after a collision. */
    

double unit_normal[3];
double unit_tangential[3];
double velocity_centroid_1_before[3];
double velocity_centroid_2_before[3];
double velocity_centroid_difference_before[3];
double tangential_top[3];
double omega1_before[3];
double omega2_before[3];
double omega1_after[3];
double omega2_after[3];
// No spinning for the moment
omega1_before[0]=0;
omega1_before[1]=0;
omega1_before[2]=0;
omega2_before[0]=0;
omega2_before[1]=0;
omega2_before[2]=0;
// No spinning for the moment
velocity_centroid_1_before[0]=velocity1_before[0];
velocity_centroid_1_before[1]=velocity1_before[1];
velocity_centroid_1_before[2]=velocity1_before[2];
velocity_centroid_2_before[0]=velocity2_before[0];
velocity_centroid_2_before[1]=velocity2_before[1];
velocity_centroid_2_before[2]=velocity2_before[2];
omega1_after[0]=0;
omega1_after[1]=0;
omega1_after[2]=0;
omega2_after[0]=0;
omega2_after[1]=0;
omega2_after[2]=0;
velocity1_after[0]=0;
velocity1_after[1]=0;
velocity1_after[2]=0;
velocity2_after[0]=0;
velocity2_after[1]=0;
velocity2_after[2]=0;

normal(position1,position2,unit_normal);

velocity_centroid_difference_before[0]=velocity_centroid_2_before[0]-velocity_centroid_1_before[0];
velocity_centroid_difference_before[1]=velocity_centroid_2_before[1]-velocity_centroid_1_before[1];
velocity_centroid_difference_before[2]=velocity_centroid_2_before[2]-velocity_centroid_1_before[2];

tangential_top[0]=velocity_centroid_difference_before[0]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[0];
tangential_top[1]=velocity_centroid_difference_before[1]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[1];
tangential_top[2]=velocity_centroid_difference_before[2]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[2];

tangential(tangential_top,unit_tangential);

/*   en and es are the coefficients of restitution
     in the normal and tangential directions. Set to 1.0 
     they indicate perfectly elastic collisions.  */
 
double en = 1.0;
double es = -1.0;

double jay[3];

jay[0] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[0]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[0];
jay[1] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[1]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[1];
jay[2] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[2]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[2];

velocity1_after[0] = velocity1_before[0]+ jay[0]/mass1; 
velocity1_after[1] = velocity1_before[1]+ jay[1]/mass1; 
velocity1_after[2] = velocity1_before[2]+ jay[2]/mass1; 

velocity2_after[0] = velocity2_before[0]- jay[0]/mass2; 
velocity2_after[1] = velocity2_before[1]- jay[1]/mass2; 
velocity2_after[2] = velocity2_before[2]- jay[2]/mass2; 

}

double fluid_velocity_calc(double position[]) 
{
/* This function computes the fluid velocity.
   In the current setting we assume the fluid velocity
   varies quadratically with respect to distance from the 
   bed of the river.   */
   double veloc; 
   veloc = position[1]*position[1]; 
   return veloc;
}

double fluid_velocity_gradient_calc(double position[]) 
{
/* This function computes the fluid velocity gradient.
   In the current setting we assume the fluid velocity gradient
   varies linearly with respect to distance from the 
   bed of the river.   */
velocity_gradient = 2.0*position[1];
return velocity_gradient;
}

double body_force(double position[], double velocity[], double give_diameter, double force_b[])
{
/* In this function we compute the particle's body force as its weight
   minus its buoyancy.   */

double beta = 0.0;

particle_volume = (1.0/6.0)*3.141*pow(give_diameter,3);

  force_b[0] = (density_particle-density_fluid)*gravity*particle_volume*sin(beta);
  force_b[1] = -1.0*(density_particle-density_fluid)*gravity*particle_volume*cos(beta);
  force_b[2] = 0.0; 

}

double drag_force(double position[], double velocity[], double give_diameter, double force_d[])
{
/* In this function we compute the drag force on the particle.  
   To do this we need to compute the flow Reynold's number and
   employ some empircal relationships. */
double kinematic_viscosity = 1.013E-6;
double dynamic_viscosity = 1.013E-3;
double cd;
double Re;
double vx, vy, vz;
double resultant_velocity;
double fluid_velocity;
fluid_velocity = fluid_velocity_calc(position);
vx = velocity[0]-fluid_velocity;
vy = velocity[1];
vz = velocity[2];
resultant_velocity = sqrt( vx*vx + vy*vy + vz*vz );

Re = abs(vx)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[0]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vx;}
else
{force_d[0]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vx;}

Re = abs(vy)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[1]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vy;}
else
{force_d[1]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vy;}

Re = abs(vz)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[2]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vz;}
else
{force_d[2]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vz;}


//force_d[0]=0.0;
//force_d[1]=0.0;
//force_d[2]=0.0;

}

double shear_lift_force(double position[], double velocity[], double give_diameter, double force_sl[])
{
/* In this function we compute the lift force on the particle due to 
   the differential in velocity above and below the particle.  */
double vx, vy, vz;
double fluid_velocity_top;
double fluid_velocity_btm;
double cl;
double vxtop, vxbtm;
double velsqtop, velsqbtm;

position[1]=position[1]+0.5*give_diameter;
fluid_velocity_top = fluid_velocity_calc(position);
position[1]=position[1]-give_diameter;
fluid_velocity_btm = fluid_velocity_calc(position);
position[1]=position[1]+0.5*give_diameter;

vxtop = velocity[0]-fluid_velocity_top;
vxbtm = velocity[0]-fluid_velocity_btm;
vy = velocity[1];
vz = velocity[2];
cl=0.8;
velsqtop = vxtop*vxtop + vy*vy + vz*vz;
velsqbtm = vxbtm*vxbtm + vy*vy + vz*vz;
force_sl[0]=0.0;
force_sl[1]=0.5*0.25*3.141*give_diameter*give_diameter*cl*density_fluid*(velsqtop-velsqbtm);
force_sl[2]=0.0;
}

double magnus_lift_force(double position[], double velocity[], double omega[], double force_ml[])
{
double vx, vy, vz;
double fluid_velocity;
fluid_velocity = fluid_velocity_calc(position);
vx = velocity[0]-fluid_velocity;
vy = velocity[1];
vz = velocity[2];
//force_ml[0]=0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*vy;
//force_ml[1]=-0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*vx;
//force_ml[2]=0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*0.0;
force_ml[0]=0.0;
force_ml[1]=0.0;
force_ml[2]=0.0;
}


void advance_one_timestep(double dt, double mass, double diameter_give, double position[], double velocity[], double omega[])
{
      int i, j;
      double force[3];
      double velocity_fluid_x;
      double velocity_gradient_fluid_x;
      double bedslope;
      double gravity;
//      double diameter_particle = 0.001;
      double volume;
      double Re;

      double distance;
      double d_force[3];
      double l_force[3];
      double b_force[3];
      double sl_force[3];
      double ml_force[3];
      double force_dashpot[3];
      double ks=0.1;
      double neta=0.1;
      double position_temp2[3];

      body_force(position,velocity,diameter_give,b_force);
      drag_force(position,velocity,diameter_give,d_force);
      shear_lift_force(position,velocity,diameter_give,sl_force);
      magnus_lift_force(position,velocity,omega,ml_force);

              // This gives us the flexibility of adding in dashpot forces at a later date
              // shoudl we wish.

              force_dashpot[0]=0.0;
              force_dashpot[1]=0.0;
              force_dashpot[2]=0.0;

 
      force[0]=b_force[0]-d_force[0]+ml_force[0]+force_dashpot[0];
      force[1]=b_force[1]-d_force[1]+ml_force[1]+force_dashpot[1];
      force[2]=b_force[2]-d_force[2]+ml_force[2]+force_dashpot[2];

      mass = density_particle*4.0*3.14*pow((diameter_give*0.5),3)/3.0;

      velocity[0]=velocity[0]+dt*(force[0]/mass);
      velocity[1]=velocity[1]+dt*(force[1]/mass);
      velocity[2]=velocity[2]+dt*(force[2]/mass);
      position[0]=position[0]+dt*velocity[0];
      position[1]=position[1]+dt*velocity[1];
      position[2]=position[2]+dt*velocity[2];


      // If a particle goes below the river bed set it on the river bed
      //  - - - Reality check !!    
      if(position[1]<0) {position[1]=0.0;}
}

void main(int argc, char *argv[])
{
double velocity_temp[3];
double velocity_temp_after[3];
double position_temp[3];
double velocity_temp2[3];
double velocity_temp2_after[3];
double position_temp2[3];
double omega_temp[3];
double force[3];
double mass[5000];
double velocity[5000][3];
double position[5000][3];
double omega[5000][3];
double distance;
int i, j;
double dt;
double mass_give;
double diameter_give;
double mass1;
double mass2;
double time;
double ewall=1.0;

  //number_of_particles = atoi(argv[1]);
  number_of_particles = 50;

  for(i=0;i<number_of_particles;++i)
  {
  particle_diameter[i]=1.1/150.0;
  position[i][0]=((double)rand()/(double)RAND_MAX)*0.3;
  position[i][1]=(double)i / (double)number_of_particles;
  position[i][2]=0.0;
  velocity[i][0]=0.0;
  velocity[i][1]=0.0;
  position[i][2]=0.0;
  velocity[i][2]=0.0;
  mass[i] = density_particle*4.0*3.14*pow((particle_diameter[i]*0.5),3)/3.0;
         allposition[i][0]=position[i][0];
         allposition[i][1]=position[i][1];
         allposition[i][2]=position[i][2];
         allvelocity[i][0]=velocity[i][0];
         allvelocity[i][1]=velocity[i][1];
         allvelocity[i][2]=velocity[i][2];
  omega[i][0]=0.0;
  omega[i][1]=0.0;
  omega[i][2]=0.0;
         allomega[i][0]=omega[i][0];
         allomega[i][1]=omega[i][1];
         allomega[i][2]=omega[i][2];
  }

   char filename[255];
   FILE *out;
   dt=0.0005;
   // Simulation time is 2.5 seconds here. You can change this.
   while(time<2.5)
   {

      sprintf(filename,"trajectory%d.csv",icount);
      icount++;
      out = fopen(filename,"w");
      fprintf(out,"i,time,d,x,y,z,vx,vy,vz\n");
        for(i=0;i<number_of_particles;++i)
        {
        //fprintf(out,"%d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",i,time,particle_diameter[i],position[i][0],position[i][1],position[i][2],velocity[i][0],velocity[i][1],velocity[i][2]);
        }

     for(i=0;i<number_of_particles;++i)
     {
         position_temp[0]=position[i][0];
         position_temp[1]=position[i][1];
         position_temp[2]=position[i][2];
         velocity_temp[0]=velocity[i][0];
         velocity_temp[1]=velocity[i][1];
         velocity_temp[2]=velocity[i][2];
         omega_temp[0]=omega[i][0];
         omega_temp[1]=omega[i][1];
         omega_temp[2]=omega[i][2];
         mass_give = mass[i];
         diameter_give = particle_diameter[i];

         advance_one_timestep(dt, mass_give, diameter_give, position_temp, velocity_temp, omega_temp);

         position[i][0]=position_temp[0];
         position[i][1]=position_temp[1];
         position[i][2]=position_temp[2];
         velocity[i][0]=velocity_temp[0];
         velocity[i][1]=velocity_temp[1];
         velocity[i][2]=velocity_temp[2];
         omega[i][0]=omega_temp[0];
         omega[i][1]=omega_temp[1];
         omega[i][2]=omega_temp[2];

         allposition[i][0]=position_temp[0];
         allposition[i][1]=position_temp[1];
         allposition[i][2]=position_temp[2];
         allvelocity[i][0]=velocity_temp[0];
         allvelocity[i][1]=velocity_temp[1];
         allvelocity[i][2]=velocity_temp[2];
         allomega[i][0]=omega_temp[0];
         allomega[i][1]=omega_temp[1];
         allomega[i][2]=omega_temp[2];
         

     }

     // Check for collisions
     // If collisions correct velocities

        for(i=0;i<number_of_particles;++i)
        {
        for(j=i;j<number_of_particles;++j)
        {

            position_temp[0]=position[i][0];
            position_temp[1]=position[i][1];
            position_temp[2]=position[i][2];
            velocity_temp[0]=velocity[i][0];
            velocity_temp[1]=velocity[i][1];
            velocity_temp[2]=velocity[i][2];
            position_temp2[0]=position[j][0];
            position_temp2[1]=position[j][1];
            position_temp2[2]=position[j][2];
            velocity_temp2[0]=velocity[j][0];
            velocity_temp2[1]=velocity[j][1];
            velocity_temp2[2]=velocity[j][2];
            mass1 = mass[i];
            mass2 = mass[j];

           distance = distance_calc(position_temp,position_temp2);
           if(distance<0.5*(particle_diameter[i]+particle_diameter[j]))
           {
              if(i!=j)
              {
              collision(mass1,mass2,position_temp,position_temp2,velocity_temp,velocity_temp2,velocity_temp_after,velocity_temp2_after);
              velocity[i][0]=velocity_temp_after[0];
              velocity[i][1]=velocity_temp_after[1];
              velocity[i][2]=velocity_temp_after[2];
              velocity[j][0]=velocity_temp2_after[0];
              velocity[j][1]=velocity_temp2_after[1];
              velocity[j][2]=velocity_temp2_after[2];
              } 
           }
           else
           {
            //  do nothing
           }
        }
         // If hits bed it bounces back
         if(position[i][1]<0.5*particle_diameter[i])
            {
            velocity[i][1]=-1.0*ewall*velocity[i][1];
            }
         // If it flows out of domain put it back in at start 
         if(position[i][0]>5.0)
            {
            position[i][0]=0.0;
            }
        }
 


       time = time + dt;
       for(i=0;i<number_of_particles;++i)
       {
       fprintf(out,"%d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",i,time,particle_diameter[i],position[i][0],position[i][1],position[i][2],velocity[i][0],velocity[i][1],velocity[i][2]);
       }
       fclose(out);
   }


}
