#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <mpi.h>

/* 
Copyright J. Russell Manson, Ph.D., P.E.
Contact me at: Russell.Manson@Stockton.edu
*/

/* Define some important variables
   that are used globally           
   Most variable names are self explanatory*/

double density_fluid = 1000.0;
double density_particle = 2650.0;
double gravity = 9.8;
double particle_diameter[1250];
double particle_mass;
double particle_volume;
double allposition[1250][3];
double allvelocity[1250][3];
double allomega[1250][3];
int icount=1;
int number_of_particles;

double effective_mass(double mass1, double mass2) 
{
/* This function is used to compute the 
   effective mass of the colliding particles. */
double mass;
return mass = mass1*mass2/(mass1+mass2);
}

double dot_product(double x[], double y[]) 
{
/* This function computes the dot product
   of two vectors.                            */
double magnitude;
return magnitude = x[0]*y[0]+x[1]*y[1]+x[2]*y[2];
}

void cross_product(double x[], double y[], double xy[]) 
{
/* This function computes the cross product
   of two vectors.                            */
xy[0] = x[1]*y[2]-x[2]*y[1];
xy[1] = -1.0*(x[0]*y[2]-x[2]*y[0]);
xy[2] = x[0]*y[1]-x[1]*y[1];
}

double norm(double x[]) 
{
/* This function computes the norm
   of a vector.                            */
double magnitude;
return magnitude = sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
}

double distance_calc(double x[], double y[]) 
{
/* This function computes the distance
   between two points given their 3d coordinates.  */
double diff[3];
double magnitude;
int i;
   for(i=0;i<3;++i)
   {
   diff[i]=y[i]-x[i];
   }
   magnitude = norm(diff);
   return magnitude;
}

void normal(double x[], double y[], double unit_normal[]) 
{
/* This function computes the unit normal vector  */
double diff[3];
double magnitude;
int i;
   for(i=0;i<3;++i)
   {
   diff[i]=y[i]-x[i];
   }
   for(i=0;i<3;++i)
   {
   unit_normal[i]=diff[i]/norm(diff);
   }
}

void tangential(double x[], double unit_tangential[]) 
{
/* This function computes the unit tangential vector  */
int i;
   for(i=0;i<3;++i)
   {
   unit_tangential[i]=x[i]/(norm(x)+1E-99);
   }
}

void collision(double mass1, double mass2, double position1[], double position2[], double velocity1_before[], double velocity2_before[], double velocity1_after[], double velocity2_after[])
{
/* Given the positions and velocities before a collision
   this function computes the velocities after a collision. */
    

double unit_normal[3];
double unit_tangential[3];
double velocity_centroid_1_before[3];
double velocity_centroid_2_before[3];
double velocity_centroid_difference_before[3];
double tangential_top[3];
double omega1_before[3];
double omega2_before[3];
double omega1_after[3];
double omega2_after[3];
// No spinning for the moment
omega1_before[0]=0;
omega1_before[1]=0;
omega1_before[2]=0;
omega2_before[0]=0;
omega2_before[1]=0;
omega2_before[2]=0;
// No spinning for the moment
velocity_centroid_1_before[0]=velocity1_before[0];
velocity_centroid_1_before[1]=velocity1_before[1];
velocity_centroid_1_before[2]=velocity1_before[2];
velocity_centroid_2_before[0]=velocity2_before[0];
velocity_centroid_2_before[1]=velocity2_before[1];
velocity_centroid_2_before[2]=velocity2_before[2];
omega1_after[0]=0;
omega1_after[1]=0;
omega1_after[2]=0;
omega2_after[0]=0;
omega2_after[1]=0;
omega2_after[2]=0;
velocity1_after[0]=0;
velocity1_after[1]=0;
velocity1_after[2]=0;
velocity2_after[0]=0;
velocity2_after[1]=0;
velocity2_after[2]=0;

normal(position1,position2,unit_normal);

velocity_centroid_difference_before[0]=velocity_centroid_2_before[0]-velocity_centroid_1_before[0];
velocity_centroid_difference_before[1]=velocity_centroid_2_before[1]-velocity_centroid_1_before[1];
velocity_centroid_difference_before[2]=velocity_centroid_2_before[2]-velocity_centroid_1_before[2];

tangential_top[0]=velocity_centroid_difference_before[0]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[0];
tangential_top[1]=velocity_centroid_difference_before[1]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[1];
tangential_top[2]=velocity_centroid_difference_before[2]-dot_product(velocity_centroid_difference_before,unit_normal)*unit_normal[2];

tangential(tangential_top,unit_tangential);

/*   en and es are the coefficients of restitution
     in the normal and tangential directions. Set to 1.0 
     they indicate perfectly elastic collisions.  */
 
double en = 1.0;
double es = -1.0;

double jay[3];

jay[0] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[0]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[0];
jay[1] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[1]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[1];
jay[2] = effective_mass(mass1,mass2) * (1.0+en) * dot_product(velocity_centroid_difference_before,unit_normal) * unit_normal[2]
       + (2.0/7.0) * effective_mass(mass1,mass2) * (1.0+es) * dot_product(velocity_centroid_difference_before,unit_tangential) * unit_tangential[2];

velocity1_after[0] = velocity1_before[0]+ jay[0]/mass1; 
velocity1_after[1] = velocity1_before[1]+ jay[1]/mass1; 
velocity1_after[2] = velocity1_before[2]+ jay[2]/mass1; 

velocity2_after[0] = velocity2_before[0]- jay[0]/mass2; 
velocity2_after[1] = velocity2_before[1]- jay[1]/mass2; 
velocity2_after[2] = velocity2_before[2]- jay[2]/mass2; 

}

double fluid_velocity_calc(double position[]) 
{
/* This function computes the fluid velocity.
   In the current setting we assume the fluid velocity
   varies quadratically with respect to distance from the 
   bed of the river.   */
   double veloc; 
   veloc = position[1]*position[1]; 
   return veloc;
}

double fluid_velocity_gradient_calc(double position[]) 
{
/* This function computes the fluid velocity gradient.
   In the current setting we assume the fluid velocity gradient
   varies linearly with respect to distance from the 
   bed of the river.   */
velocity_gradient = 2.0*position[1];
return velocity_gradient;
}

double body_force(double position[], double velocity[], double give_diameter, double force_b[])
{
/* In this function we compute the particle's body force as its weight
   minus its buoyancy.   */

double beta = 0.0;

particle_volume = (1.0/6.0)*3.141*pow(give_diameter,3);

  force_b[0] = (density_particle-density_fluid)*gravity*particle_volume*sin(beta);
  force_b[1] = -1.0*(density_particle-density_fluid)*gravity*particle_volume*cos(beta);
  force_b[2] = 0.0; 

}

double drag_force(double position[], double velocity[], double give_diameter, double force_d[])
{
/* In this function we compute the drag force on the particle.  
   To do this we need to compute the flow Reynold's number and
   employ some empircal relationships. */
double kinematic_viscosity = 1.013E-6;
double dynamic_viscosity = 1.013E-3;
double cd;
double Re;
double vx, vy, vz;
double resultant_velocity;
double fluid_velocity;
fluid_velocity = fluid_velocity_calc(position);
vx = velocity[0]-fluid_velocity;
vy = velocity[1];
vz = velocity[2];
resultant_velocity = sqrt( vx*vx + vy*vy + vz*vz );

Re = abs(vx)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[0]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vx;}
else
{force_d[0]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vx;}

Re = abs(vy)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[1]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vy;}
else
{force_d[1]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vy;}

Re = abs(vz)*give_diameter/kinematic_viscosity;
if(Re>1000.0)
{cd=0.44;}
else
{cd=(24.0/(Re+1E-99))*(1.0+0.14*pow(Re,0.7));}
if(Re<1)
{force_d[2]=6.0*3.141*dynamic_viscosity*(give_diameter/2.0)*vz;}
else
{force_d[2]=cd*3.141*0.25*give_diameter*give_diameter*density_fluid*resultant_velocity*vz;}


//force_d[0]=0.0;
//force_d[1]=0.0;
//force_d[2]=0.0;

}

double shear_lift_force(double position[], double velocity[], double give_diameter, double force_sl[])
{
/* In this function we compute the lift force on the particle due to 
   the differential in velocity above and below the particle.  */
double vx, vy, vz;
double fluid_velocity_top;
double fluid_velocity_btm;
double cl;
double vxtop, vxbtm;
double velsqtop, velsqbtm;

position[1]=position[1]+0.5*give_diameter;
fluid_velocity_top = fluid_velocity_calc(position);
position[1]=position[1]-give_diameter;
fluid_velocity_btm = fluid_velocity_calc(position);
position[1]=position[1]+0.5*give_diameter;

vxtop = velocity[0]-fluid_velocity_top;
vxbtm = velocity[0]-fluid_velocity_btm;
vy = velocity[1];
vz = velocity[2];
cl=0.8;
velsqtop = vxtop*vxtop + vy*vy + vz*vz;
velsqbtm = vxbtm*vxbtm + vy*vy + vz*vz;
force_sl[0]=0.0;
force_sl[1]=0.5*0.25*3.141*give_diameter*give_diameter*cl*density_fluid*(velsqtop-velsqbtm);
force_sl[2]=0.0;
}

double magnus_lift_force(double position[], double velocity[], double omega[], double force_ml[])
{
double vx, vy, vz;
double fluid_velocity;
fluid_velocity = fluid_velocity_calc(position);
vx = velocity[0]-fluid_velocity;
vy = velocity[1];
vz = velocity[2];
//force_ml[0]=0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*vy;
//force_ml[1]=-0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*vx;
//force_ml[2]=0.125*3.141*pow(particle_diameter,3)*density_fluid*(omega[1]-0.5*fluid_velocity_gradient_calc(position))*0.0;
force_ml[0]=0.0;
force_ml[1]=0.0;
force_ml[2]=0.0;
}


void advance_one_timestep(double dt, double mass, double diameter_give, double position[], double velocity[], double omega[])
{
      int i, j;
      double force[3];
      double velocity_fluid_x;
      double velocity_gradient_fluid_x;
      double bedslope;
      double gravity;
//      double diameter_particle = 0.001;
      double volume;
      double Re;

      double distance;
      double d_force[3];
      double l_force[3];
      double b_force[3];
      double sl_force[3];
      double ml_force[3];
      double force_dashpot[3];
      double ks=0.1;
      double neta=0.1;
      double position_temp2[3];

      body_force(position,velocity,diameter_give,b_force);
      drag_force(position,velocity,diameter_give,d_force);
      shear_lift_force(position,velocity,diameter_give,sl_force);
      magnus_lift_force(position,velocity,omega,ml_force);

              // This gives us the flexibility of adding in dashpot forces at a later date
              // shoudl we wish.

              force_dashpot[0]=0.0;
              force_dashpot[1]=0.0;
              force_dashpot[2]=0.0;

 
      force[0]=b_force[0]-d_force[0]+ml_force[0]+force_dashpot[0];
      force[1]=b_force[1]-d_force[1]+ml_force[1]+force_dashpot[1];
      force[2]=b_force[2]-d_force[2]+ml_force[2]+force_dashpot[2];

      mass = density_particle*4.0*3.14*pow((diameter_give*0.5),3)/3.0;

      velocity[0]=velocity[0]+dt*(force[0]/mass);
      velocity[1]=velocity[1]+dt*(force[1]/mass);
      velocity[2]=velocity[2]+dt*(force[2]/mass);
      position[0]=position[0]+dt*velocity[0];
      position[1]=position[1]+dt*velocity[1];
      position[2]=position[2]+dt*velocity[2];


      // If a particle goes below the river bed set it on the river bed
      //  - - - Reality check !!    
      if(position[1]<0) {position[1]=0.0;}
}

void main(int argc, char *argv[])
{
double velocity_temp[3];
double velocity_temp_after[3];
double position_temp[3];
double velocity_temp2[3];
double velocity_temp2_after[3];
double position_temp2[3];
double omega_temp[3];
double force[3];
double mass[1250];
double velocity[1250][3];
double position[1250][3];
double omega[1250][3];
double distance;
int i, j;
double dt;
double mass_give;
double diameter_give;
double mass1;
double mass2;
double time;
double ewall=1.0;

double xposition[1250];
double yposition[1250];
double zposition[1250];
double uvelocity[1250];
double vvelocity[1250];
double wvelocity[1250];
double xpositionproc[1250];
double ypositionproc[1250];
double zpositionproc[1250];
double uvelocityproc[1250];
double vvelocityproc[1250];
double wvelocityproc[1250];

int numprocs;
int rank;
int particles_per_process;
int start_compute;
int end_compute;

MPI_Status status;

MPI_Init(&argc, &argv);
MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  // Note that total particle number is read in from terminal at run time

  number_of_particles = atoi(argv[1]);

  particles_per_process = number_of_particles / numprocs;

  start_compute = rank*particles_per_process; 
  end_compute = start_compute + particles_per_process - 1; 

     ///printf("%d %d %d\n",rank,start_compute,end_compute);

  if(rank==0)
  {
  for(i=0;i<number_of_particles;++i)
  {
  // Initialize particle diameters and positions employing random numbers
  particle_diameter[i]=(((double)rand()/(double)RAND_MAX)*4.0+10.0)/150.0;
  position[i][0]=((double)rand()/(double)RAND_MAX)*0.3;
  position[i][1]=(double)i / (double)number_of_particles;
  position[i][2]=0.0;
  velocity[i][0]=0.0;
  velocity[i][1]=0.0;
  velocity[i][2]=0.0;
  omega[i][0]=0.0;
  omega[i][1]=0.0;
  omega[i][2]=0.0;
  mass[i] = density_particle*4.0*3.14*pow((particle_diameter[i]*0.5),3)/3.0;
  }
  }

       dt=0.00001;
       FILE *out;
   // Maximum simulation time is 0.3 seconds. You can change that here.
   while(time<0.3)
   {

      if(rank==0)
      {
      // Note that we write out results for each timestep to separate files.
      // This allows us to do the paraview visualization easily.
      char filename[255];
      sprintf(filename,"trajectory%d.csv",icount);
      icount++;
      out = fopen(filename,"w");
      fprintf(out,"i,time,d,x,y,z,vx,vy,vz\n");
        for(i=0;i<number_of_particles;++i)
        {
        fprintf(out,"%d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",i,time,particle_diameter[i],position[i][0],position[i][1],position[i][2],velocity[i][0],velocity[i][1],velocity[i][2]);
        }
      }


        if(rank==0)
        {

           for(i=0;i<number_of_particles;++i)
           {
           xposition[i]=position[i][0];
           yposition[i]=position[i][1];
           zposition[i]=position[i][2];
           uvelocity[i]=velocity[i][0];
           vvelocity[i]=velocity[i][1];
           wvelocity[i]=velocity[i][2];
           }


           // Send all particle information to all cores

           for(i=1;i<numprocs;++i)
           {
           MPI_Send(xposition,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(yposition,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(zposition,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(uvelocity,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(vvelocity,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(wvelocity,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(particle_diameter,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           MPI_Send(mass,number_of_particles,MPI_DOUBLE,i,99,MPI_COMM_WORLD);
           }

         }

        if(rank!=0)
        {
           // Receive all particle information by all cores
           MPI_Recv(xposition,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(yposition,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(zposition,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(uvelocity,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(vvelocity,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(wvelocity,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(particle_diameter,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
           MPI_Recv(mass,number_of_particles,MPI_DOUBLE,0,99,MPI_COMM_WORLD,&status);
        }


        start_compute = rank*particles_per_process; 
        end_compute = start_compute + particles_per_process - 1; 

     // Perform timestep advancement on all particles within the core

     for(i=start_compute;i<end_compute+1;++i)
     {
         position_temp[0]=xposition[i];
         position_temp[1]=yposition[i];
         position_temp[2]=zposition[i];
         velocity_temp[0]=uvelocity[i];
         velocity_temp[1]=vvelocity[i];
         velocity_temp[2]=wvelocity[i];
         omega_temp[0]=0.0;
         omega_temp[1]=0.0;
         omega_temp[2]=0.0;
         mass_give = mass[i];
         diameter_give = particle_diameter[i];

         advance_one_timestep(dt, mass_give, diameter_give, position_temp, velocity_temp, omega_temp);

         xposition[i]=position_temp[0];
         yposition[i]=position_temp[1];
         zposition[i]=position_temp[2];
         uvelocity[i]=velocity_temp[0];
         vvelocity[i]=velocity_temp[1];
         wvelocity[i]=velocity_temp[2];

     }

     // Check for collisions
     // If collisions occur correct velocities

        if(1==1)
        {

        for(i=start_compute;i<end_compute+1;++i)
        {
        for(j=i;j<number_of_particles;++j)
        {

            position_temp[0]=xposition[i];
            position_temp[1]=yposition[i];
            position_temp[2]=zposition[i];
            velocity_temp[0]=uvelocity[i];
            velocity_temp[1]=vvelocity[i];
            velocity_temp[2]=wvelocity[i];
            position_temp2[0]=xposition[j];
            position_temp2[1]=yposition[j];
            position_temp2[2]=zposition[j];
            velocity_temp2[0]=uvelocity[j];
            velocity_temp2[1]=vvelocity[j];
            velocity_temp2[2]=wvelocity[j];
            mass1 = mass[i];
            mass2 = mass[j];

           distance = distance_calc(position_temp,position_temp2);

           if(distance<0.5*(particle_diameter[i]+particle_diameter[j]))
           {
              if(i!=j)
              {
              //Collision
              //printf("Collision distance = %lf\n",distance);
              //printf("%lf %lf %lf %lf %lf %lf\n",velocity_temp[0],velocity_temp[1],velocity_temp[2],velocity_temp2[0],velocity_temp2[1],velocity_temp2[2]);
              collision(mass1,mass2,position_temp,position_temp2,velocity_temp,velocity_temp2,velocity_temp_after,velocity_temp2_after);
              //printf("%lf %lf %lf %lf %lf %lf\n",velocity_temp_after[0],velocity_temp_after[1],velocity_temp_after[2],velocity_temp2_after[0],velocity_temp2_after[1],velocity_temp2_after[2]);
              ////position[i][0]=position_temp[0];
              ////position[i][1]=position_temp[1];
              ////position[i][2]=position_temp[1];
              uvelocity[i]=velocity_temp_after[0];
              vvelocity[i]=velocity_temp_after[1];
              wvelocity[i]=velocity_temp_after[2];
              ////position[j][0]=position_temp2[0];
              ////position[j][1]=position_temp2[0];
              ////position[j][2]=position_temp2[0];
              uvelocity[j]=velocity_temp2_after[0];
              vvelocity[j]=velocity_temp2_after[1];
              wvelocity[j]=velocity_temp2_after[2];
              } 
           }
           else
           {
            //velocity[i][0]=velocity_temp[0];
            //velocity[i][1]=velocity_temp[1];
            //velocity[i][2]=velocity_temp[2];
            //velocity[j][0]=velocity_temp2[0];
            //velocity[j][1]=velocity_temp2[1];
            //velocity[j][2]=velocity_temp2[2];
           }
        }

         // If particle hits bed it bounces back
         if(yposition[i]<0.5*particle_diameter[i] && vvelocity[i]<0.0)
            {
            vvelocity[i]=-1.0*ewall*vvelocity[i];
            }
         // If particle flows out of domain then 
         // put it back to start.
         if(xposition[i]>2.0)
            {
            xposition[i]=0.0;
            }
        }


}
////       End collisions
 
        int icounter=0;
        for(i=start_compute;i<end_compute+1;++i)
        {
        xpositionproc[icounter]=xposition[i];
        ypositionproc[icounter]=yposition[i];
        zpositionproc[icounter]=zposition[i];
        uvelocityproc[icounter]=uvelocity[i];
        vvelocityproc[icounter]=vvelocity[i];
        wvelocityproc[icounter]=wvelocity[i];
        ++icounter;
        }

       time = time + dt;

        if(rank!=0)
        {
        // Send results back to rank 0 for book-keeping
        MPI_Send(xpositionproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        MPI_Send(ypositionproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        MPI_Send(zpositionproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        MPI_Send(uvelocityproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        MPI_Send(vvelocityproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        MPI_Send(wvelocityproc,particles_per_process,MPI_DOUBLE,0,99,MPI_COMM_WORLD);
        }

        if(rank==0)
        {
        icounter=0;
        start_compute = 0;
        end_compute = start_compute + particles_per_process - 1; 
        for(j=start_compute;j<end_compute+1;++j)
        {
        position[j][0]=xpositionproc[icounter];
        position[j][1]=ypositionproc[icounter];
        position[j][2]=zpositionproc[icounter];
        velocity[j][0]=uvelocityproc[icounter];
        velocity[j][1]=vvelocityproc[icounter];
        velocity[j][2]=wvelocityproc[icounter];
        ++icounter;
        }

        for(i=1;i<numprocs;++i)
        {
        // Receive results on rank 0
        MPI_Recv(xpositionproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        MPI_Recv(ypositionproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        MPI_Recv(zpositionproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        MPI_Recv(uvelocityproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        MPI_Recv(vvelocityproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        MPI_Recv(wvelocityproc,particles_per_process,MPI_DOUBLE,i,99,MPI_COMM_WORLD,&status);
        icounter=0;
        start_compute = i*particles_per_process; 
        end_compute = start_compute + particles_per_process - 1; 
        for(j=start_compute;j<end_compute+1;++j)
        {
        position[j][0]=xpositionproc[icounter];
        position[j][1]=ypositionproc[icounter];
        position[j][2]=zpositionproc[icounter];
        velocity[j][0]=uvelocityproc[icounter];
        velocity[j][1]=vvelocityproc[icounter];
        velocity[j][2]=wvelocityproc[icounter];
        ++icounter;
        }
        }


        }



       if(rank==0)
       { 
          // Print out results
          for(i=0;i<number_of_particles;++i)
          {
          fprintf(out,"%d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",i,time,particle_diameter[i],position[i][0],position[i][1],position[i][2],velocity[i][0],velocity[i][1],velocity[i][2]);
          }
          fclose(out);
       }
   }


MPI_Finalize();
}
